/****************************************************************************
*
*    Copyright (c) 2019 Vivante Corporation
*
*    Permission is hereby granted, free of charge, to any person obtaining a
*    copy of this software and associated documentation files (the "Software"),
*    to deal in the Software without restriction, including without limitation
*    the rights to use, copy, modify, merge, publish, distribute, sublicense,
*    and/or sell copies of the Software, and to permit persons to whom the
*    Software is furnished to do so, subject to the following conditions:
*
*    The above copyright notice and this permission notice shall be included in
*    all copies or substantial portions of the Software.
*
*    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
*    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
*    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
*    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
*    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
*    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
*    DEALINGS IN THE SOFTWARE.
*
*****************************************************************************/


#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "vsi_nn_types.h"
#include "vsi_nn_tensor.h"
#include "vsi_nn_graph.h"
#include "vsi_nn_log.h"
#include "vsi_nn_prv.h"
#include "vsi_nn_tensor_util.h"
#include "vsi_nn_error.h"
#include "utils/vsi_nn_util.h"
#include "kernel/vsi_nn_kernel.h"

__BEGIN_DECLS

/*
 * Define kernel meta.
 */
#define _CPU_ARG_NUM            (1)
#define _CPU_INPUT_NUM          (2)
#define _CPU_OUTPUT_NUM         (1)
#define _CPU_IO_NUM             (_CPU_INPUT_NUM + _CPU_OUTPUT_NUM)
#define _CPU_PARAM_NUM          (_CPU_ARG_NUM + _CPU_IO_NUM)
#define _KERNEL_NAME            CVIVANTE_NAMESPACE("cpu.repeat")

DEF_KERNEL_EXECUTOR(_repeat_exec)
    (
    vsi_nn_kernel_node_t node,
    const vsi_nn_kernel_node_param_t * param,
    size_t param_size
    )
{
    vsi_status status = VSI_FAILURE;
    vsi_nn_kernel_tensor_t tensors[_CPU_IO_NUM] = { NULL };
    float * buffer[_CPU_IO_NUM] = { NULL };
    size_t out_elements = 0;
    vsi_nn_kernel_tensor_attr_t * attr[_CPU_IO_NUM] = { NULL };
    vsi_ssize_t i = 0, j = 0, b = 0, c = 0;
    int32_t axis = 0;
    vsi_ssize_t outerSize = 1;
    vsi_ssize_t outIdx = 0;
    vsi_ssize_t width = 0, height = 0, channel = 0, batch = 0;
    vsi_ssize_t spatial = 0, vol = 0;

    tensors[0]  = (vsi_nn_kernel_tensor_t)param[0];
    tensors[1]  = (vsi_nn_kernel_tensor_t)param[1];
    tensors[2]  = (vsi_nn_kernel_tensor_t)param[2];

    attr[0] = vsi_nn_kernel_tensor_attr_create( tensors[0] );
    CHECK_PTR_FAIL_GOTO( attr[0], "Create tensor attr buffer fail.", final );
    attr[1] = vsi_nn_kernel_tensor_attr_create( tensors[1] );
    CHECK_PTR_FAIL_GOTO( attr[1], "Create tensor attr buffer fail.", final );
    attr[2] = vsi_nn_kernel_tensor_attr_create( tensors[2] );
    CHECK_PTR_FAIL_GOTO( attr[2], "Create tensor attr buffer fail.", final );

    out_elements = vsi_nn_kernel_tensor_attr_get_size( attr[2] );

    status = vsi_nn_kernel_scalar_read_int32((vsi_nn_kernel_scalar_t)param[3], &axis);
    CHECK_STATUS_FAIL_GOTO(status, final );

    buffer[0] = (float*)vsi_nn_kernel_tensor_create_buffer( tensors[0], attr[0], TRUE );
    CHECK_PTR_FAIL_GOTO( buffer[0], "Create input0 buffer fail.", final );

    buffer[1] = (float*)vsi_nn_kernel_tensor_create_buffer( tensors[1], attr[1], TRUE );
    CHECK_PTR_FAIL_GOTO( buffer[1], "Create input0 buffer fail.", final );

    buffer[2] = (float *)malloc( out_elements * sizeof(float) );
    CHECK_PTR_FAIL_GOTO( buffer[2], "Create output buffer fail.", final );
    memset( buffer[2], 0, out_elements * sizeof(float) );

    width   = attr[0]->shape->data[0];
    height  = attr[0]->shape->data[1];
    channel = attr[0]->shape->size > 2 ? attr[0]->shape->data[2] : 1;
    batch   = attr[0]->shape->size > 3 ? attr[0]->shape->data[3] : 1;
    spatial = width * height;
    vol     = spatial * channel;

    for(i = 1; i < (int32_t)attr[0]->shape->size; i++)
    {
        outerSize *= attr[0]->shape->data[i];
    }

    if (axis == 0 && outerSize == 1)
    {
        for(i = 0; i < width; i++)
        {
            float data = buffer[0][i];
            int32_t len = (int32_t)buffer[1][i];
            for(j = 0; j < len; j++)
            {
                buffer[2][outIdx] = data;
            }
        }
    }
    else if (axis == 0)
    {
        for(b = 0; b < batch; b++)
        {
            for(c = 0; c < channel; c++)
            {
                for(i = 0; i < height; i++)
                {
                    vsi_ssize_t len = (int32_t)buffer[1][i];
                    vsi_ssize_t offset = i * width + c * spatial + b * vol;
                    for(j = 0; j < len; j++)
                    {
                        memcpy(buffer[2] + outIdx, buffer[0] + offset, sizeof(float) * width);
                        outIdx += width;
                    }
                }
            }
        }
    }
    else if (axis == 1)
    {
        for(b = 0; b < batch; b++)
        {
            for(c = 0; c < channel; c++)
            {
                for(i = 0; i < height; i++)
                {
                    vsi_ssize_t offset = i * width + c * spatial + b * vol;
                    for(j = 0; j < width; j++)
                    {
                        vsi_ssize_t len = (vsi_ssize_t)buffer[1][j];
                        float data = buffer[0][offset + j];
                        vsi_ssize_t k = 0;
                        for(k = 0; k < len; k++)
                        {
                            buffer[2][outIdx++] = data;
                        }
                    }
                }
            }
        }
    }
    else if (axis == 2)
    {
        for(b = 0; b < batch; b++)
        {
            for(c = 0; c < channel; c++)
            {
                vsi_ssize_t len = (vsi_ssize_t)buffer[1][c];
                vsi_ssize_t offset = c * spatial + b * vol;

                for(j = 0; j < len; j++)
                {
                    memcpy(buffer[2] + outIdx, buffer[0] + offset, sizeof(float) * spatial);
                    outIdx += spatial;
                }
            }
        }
    }
    else
    {
        VSILOGE("axis is not support");
        status = VSI_FAILURE;
        goto final;
    }

    status = vsi_nn_kernel_tensor_write_from_float( tensors[2], attr[2],
            buffer[2], out_elements );
    CHECK_STATUS_FAIL_GOTO( status, final );

final:
    for( i = 0; i < _CPU_IO_NUM; i ++ )
    {
        if( buffer[i] )
        {
            free( buffer[i] );
        }
    }
    for( i = 0; i < _CPU_IO_NUM; i ++ )
    {
        if(attr[i]) { vsi_nn_kernel_tensor_attr_release( &attr[i] ); }
    }
    return status;
} /* _repeat_exec() */
/*
 * Kernel params
 */
static vx_param_description_t _repeat_kernel_param_def[] =
{
    {VX_INPUT, VX_TYPE_TENSOR, VX_PARAMETER_STATE_REQUIRED},
    {VX_INPUT, VX_TYPE_TENSOR, VX_PARAMETER_STATE_REQUIRED},
    {VX_OUTPUT, VX_TYPE_TENSOR, VX_PARAMETER_STATE_REQUIRED},
    {VX_INPUT, VX_TYPE_SCALAR, VX_PARAMETER_STATE_REQUIRED},
    // Add kererl parameters here
};
#define _REPEAT_PARAM_NUM  _cnt_of_array( _repeat_kernel_param_def )

static vsi_status _query_kernel
    (
    vsi_nn_tensor_t* const* const inputs,
    vsi_nn_tensor_t* const* const outputs,
    vsi_nn_kernel_t* kernel
    )
{
    snprintf( kernel->info.name, VX_MAX_KERNEL_NAME, "%s",  _KERNEL_NAME );
    kernel->info.function    = _repeat_exec;
    kernel->info.parameters  = _repeat_kernel_param_def;
    kernel->info.numParams   = _cnt_of_array( _repeat_kernel_param_def );

    return VSI_SUCCESS;
} /* _query_kernel() */

static vsi_nn_kernel_node_t _setup
    (
    vsi_nn_graph_t              * graph,
    vsi_nn_tensor_t            ** inputs,
    size_t                        input_num,
    vsi_nn_tensor_t            ** outputs,
    size_t                        output_num,
    const vsi_nn_kernel_param_t * params,
    vsi_nn_kernel_t             * kernel
    )
{
    vsi_status status = VSI_FAILURE;
    vsi_nn_kernel_node_param_t backend_params[_CPU_PARAM_NUM] = {NULL};
    vsi_nn_kernel_node_t node = NULL;
    int32_t axis  = vsi_nn_kernel_param_get_int32( params, "axis" );

    status = _query_kernel( inputs, outputs, kernel );
    if( VSI_SUCCESS == status)
    {
        node = vsi_nn_kernel_create_node( graph, kernel );
        if( node )
        {
            /* Set inputs and outputs */
            vsi_nn_kernel_node_pack_io( backend_params, _CPU_PARAM_NUM,
                    inputs, _CPU_INPUT_NUM, outputs, _CPU_OUTPUT_NUM );
            backend_params[3] = vsi_nn_kernel_scalar_create( graph, I32, &axis );

            /* Pass parameters to node. */
            status = vsi_nn_kernel_node_pass_param( node, backend_params, _CPU_PARAM_NUM );
            CHECK_STATUS( status );
            vsi_nn_kernel_scalar_release( &backend_params[3] );
        }
        else
        {
            status = VSI_FAILURE;
        }
    }
    return node;
} /* _setup() */

__END_DECLS

REGISTER_BACKEND_CPU( repeat, _setup )
